/*

*/

package renderer.scene;

import java.util.List;
import java.util.ArrayList;
import java.awt.Color;
import java.util.Random;

/**
   A {@code Model} data structure represents a distinct geometric object
   in a {@link Scene}. A {@code Model} data structure is mainly a {@link List}
   of {@link Vertex} object and another list of {@link LineSegment} objects.
   Each {@link LineSegment} object contains two integers that are the indices
   of two {@link Vertex} objects from the {@code Models}'s vertex list. The
   two {@link Vertex} objects contain the coordinates, in the camera coordinate
   system, of the line segment's two endpoints.
<p>
   A {@code Model} represent the geometric object as a "wire-frame" of line
   segments, that is, the geometric object is drawn as a collection of "edges".
   This is a fairly simplistic way of doing 3D graphics and we will
   improve this in later renderers.
<p>
   See
<br> <a href="http://en.wikipedia.org/wiki/Wire-frame_model" target="_top">
              http://en.wikipedia.org/wiki/Wire-frame_model</a>
<br>or
<br> <a href="https://www.google.com/search?q=graphics+wireframe&tbm=isch" target="_top">
              https://www.google.com/search?q=graphics+wireframe&tbm=isch</a>
*/
public class Model
{
   public List<Vertex> vertexList = new ArrayList<>();
   public List<LineSegment> lineSegmentList = new ArrayList<>();
   public List<Color> colorList = new ArrayList<>();

   public boolean visible;


   /**
      Construct an empty {@code Model} object.
   */
   public Model()
   {
      visible = true;
   }


   /**
      Construct an (almost) empty {@code Model} object with the given default {@link Color}.

      @param c  default {@link Color} for this {@code Model}
   */
   public Model(Color c)
   {
      visible = true;
      colorList.add(c);  // default color
   }


   /**
      A "copy constructor". This constructor should make a deep copy
      of the given {@code Model}'s {@link Vertex} list and
      {@link LineSegment} list.

      @param model  {@code Model} to make a copy of
   */
   public Model(Model model) // a "copy constructor"
   {
      super();

      this.visible = model.visible;
      for (Vertex v : model.vertexList)
      {
         this.vertexList.add(new Vertex(v)); // deep copy of each Vertex
      }
      for (LineSegment ls : model.lineSegmentList)
      {
         this.lineSegmentList.add(new LineSegment(ls)); // deep copy of each LineSgement
      }
      for (Color c : model.colorList)
      {
         this.colorList.add(c); // Color objects are immutable
      }
   }


   /**
      Add a {@link Vertex} (or vertices) to this {@code Model}'s
      {@link List} of vertices.

      @param vArray  array of {@link Vertex} objects to add to this {@code Model}
   */
   public void addVertex(Vertex... vArray)
   {
      for (Vertex v : vArray)
      {
         this.vertexList.add(v);
      }
   }


   /**
      Add a {@link Color} (or colors) to this {@code Model}'s
      {@link List} of colors.

      @param cArray  array of {@link Color} objects to add to this {@code Model}
   */
   public void addColor(Color... cArray)
   {
      for (Color c : cArray)
      {
         this.colorList.add(c);
      }
   }


   /**
      Add a {@link LineSegment} to this {@code Model}'s {@link List} of
      line segments.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list. This method assumes that the appropriate vertices
      have been added to the {@code Model}'s {@link Vertex} list.

      @param i0  1st index of {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param i1  2nd index of {@link Vertex} of {@link LineSegment} to add to this {@code Model}
   */
   public void addLineSegment(int i0, int i1)
   {
      this.lineSegmentList.add(new LineSegment(i0, i1));
   }


   /**
      Add a {@link LineSegment} to this {@code Model}'s {@link List} of line
      segments. Give this line segment the given indexed {@link Color}s.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list nor does it add any color objects to the model's
      {@link Color} list. This method assumes that the appropriate vertices
      and colors have been added to the {@code Model}'s {@link Vertex} and
      (@link Color} lists.

      @param i0  1st index of {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param i1  2nd index of {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param c0  1st index of {@link Color} of {@link LineSegment} to add to this {@code Model}
      @param c1  2nd index of {@link Color} of {@link LineSegment} to add to this {@code Model}
   */
   public void addLineSegment(int i0, int i1, int c0, int c1)
   {
      this.lineSegmentList.add(new LineSegment(i0, i1, c0, c1));
   }


   /**
      Add a {@link LineSegment} to this {@code Model}'s {@link List} of line
      segments. Give this line segment the given indexed {@link Color}.
      <p>
      NOTE: This method does not add any vertices to the {@code Model}'s
      {@link Vertex} list nor does it add a color object to the model's
      {@link Color} list. This method assumes that the appropriate vertices
      and color have been added to the {@code Model}'s {@link Vertex} and
      (@link Color} lists.

      @param i0  1st index of {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param i1  2nd index of {@link Vertex} of {@link LineSegment} to add to this {@code Model}
      @param c   index of {@link Color} of {@link LineSegment} to add to this {@code Model}
   */
   public void addLineSegment(int i0, int i1, int c)
   {
      this.lineSegmentList.add(new LineSegment(i0, i1, c, c));
   }


   /**
      Set each {@link Color} in this {@code Model}'s color list
      to the same {@link Color}.

      @param c  {@link Color} for all of this model's {@link Vertex} objects
   */
   public void setColor(Color c)
   {
      if (this.colorList.isEmpty())
      {
         this.colorList.add(c);
      }
      else
      {
         for (int i = 0; i < this.colorList.size(); i++)
         {
            this.colorList.set(i, c);
         }
      }
   }


   /**
      Set each {@link Color} in this {@code Model}'s color list
      to the same random {@link Color}.
   */
   public void setRandomColor()
   {
      Random generator = new Random();
      float r = generator.nextFloat();
      float g = generator.nextFloat();
      float b = generator.nextFloat();
      Color c = new Color(r, g, b);
      setColor(c);
   }


   /**
      Set each {@link Color} in this {@code Model} color list
      to a different random {@link Color}.
   */
   public void setRandomColors()
   {
      if (this.colorList.isEmpty())
      {
         setRandomColor();
      }
      else
      {
         Random generator = new Random();
         for (int i = 0; i < this.colorList.size(); i++)
         {
            float r = generator.nextFloat();
            float g = generator.nextFloat();
            float b = generator.nextFloat();
            Color c = new Color(r, g, b);
            this.colorList.set(i, c);
         }
      }
   }


   /**
      Set each {@link Vertex} in this {@code Model}
      to a different random {@link Color}.
      <p>
      NOTE: This will destroy whatever "color structure"
      the model might possess.
   */
   public void setRandomVertexColors()
   {
      colorList = new ArrayList<Color>();
      Random generator = new Random();
      for (int i = 0; i < this.vertexList.size(); i++)
      {
         float r = generator.nextFloat();
         float g = generator.nextFloat();
         float b = generator.nextFloat();
         Color c = new Color(r, g, b);
         this.colorList.add(c);
      }
      for (LineSegment ls : this.lineSegmentList)
      {
         ls.cIndex[0] = ls.vIndex[0];
         ls.cIndex[1] = ls.vIndex[1];
      }
   }


   /**
      Set each {@link LineSegment} in this {@code Model}
      to a different (uniform) random {@link Color}.
      <p>
      NOTE: This will destroy whatever "color structure"
      the model might possess.
   */
   public void setRandomLineSegmentColors()
   {
      colorList = new ArrayList<Color>();
      Random generator = new Random();
      for (int i = 0; i < this.lineSegmentList.size(); i++)
      {
         float r = generator.nextFloat();
         float g = generator.nextFloat();
         float b = generator.nextFloat();
         Color c = new Color(r, g, b);
         this.colorList.add(c);
      }
      int i = 0;
      for (LineSegment ls : this.lineSegmentList)
      {
         ls.cIndex[0] = i;
         ls.cIndex[1] = i++;
      }
   }


   /**
      Set each endpoint of each {@link LineSegment} in this {@code Model}
      to a different random {@link Color}.
      <p>
      NOTE: This will destroy whatever "color structure"
      the model might possess.
   */
   public void setRainbowLineSegmentColors()
   {
      colorList = new ArrayList<Color>();
      Random generator = new Random();
      for (int i = 0; i < 2 * this.lineSegmentList.size(); i++)
      {
         float r = generator.nextFloat();
         float g = generator.nextFloat();
         float b = generator.nextFloat();
         Color c = new Color(r, g, b);
         this.colorList.add(c);
      }
      int i = 0;
      for (LineSegment ls : this.lineSegmentList)
      {
         ls.cIndex[0] = i++;
         ls.cIndex[1] = i++;
      }
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Model} object
   */
   @Override
   public String toString()
   {
      String result = "";
      result += "Model has " + vertexList.size() + " vertices.\n";
      result += "Model has " + colorList.size() + " colors.\n";
      result += "Model has " + lineSegmentList.size() + " line segments.\n";
      int i = 0;
      for (Vertex v : this.vertexList)
      {
         result += i + ": " + v.toString();
         i++;
      }
      //result = "Printing out this Model's " + colortList.size() + " colors:\n";
      i = 0;
      for (Color c : this.colorList)
      {
         result += i + ": " + c.toString();
         i++;
      }
      //result = "Printing out this Model's " + lineSegmentList.size() + " Line segments:\n";
      for (LineSegment ls : this.lineSegmentList)
      {
         result += ls.toString();
      }
      //result += "Done printing out Model\n";
      return result;
   }
}
